/*
 * Copyright (c) 2003, 2004 Jussi Koskela
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package fi.iki.hsivonen.webcache.dependency;

import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

public class InvalidatorThread extends Thread {    

    /** Set of queued invalidation keys */
    private LinkedHashSet invalidationSet;

    /** Boolean value indicating if the thread should stop running */    
    private volatile boolean stop;

    /** URL of the cache server */
    private URL server;

    /** Boolean value indicating if the cache is to be reseted */
    private boolean reset;

    /** Invalidation key delimiter */
    private static final String DELIMITER = "\n";

    /** Cache invalidation message */
    private static final String INVALIDATE_ALL = "*";

    /** Maximum number of queued invalidation keys */
    private static final int MAX_QUEUE_SIZE = 10000;

    /** Delay between first invalidation and posting it */
    private static final int INVALIDATION_DELAY = 1000;
    
    /** 
     * Class constructor.
     * @param url the URL of the cache server
     */
    public InvalidatorThread(URL url) {
        invalidationSet = new LinkedHashSet();
        server = url;
        reset();
    }

    /**
     * Invalidates the contents which depend on the given invalidation keys.
     * @param invKeys the set of invalidation keys
     */
    public void invalidate(Collection invKeys) {
        synchronized(invalidationSet) {
            if (invalidationSet.size() + invKeys.size() > MAX_QUEUE_SIZE) {
                reset();
            } else if (!reset) {
                invalidationSet.addAll(invKeys);
                invalidationSet.notifyAll();
            }
        }
    }

    /**
     * Invalidates the cache.
     */
    public void reset() {
        synchronized(invalidationSet) {
            reset = true;
            invalidationSet.clear();
            invalidationSet.add(INVALIDATE_ALL);
            invalidationSet.notifyAll();
        }
    }

    /**
     * Posts the invalidations to the cache server
     * @param invKeys a collection of invalidation keys
     * @return boolean value indicating the success of operation
     */
    private boolean post(Collection invKeys) {
        boolean success = false;
 
        CacheInvalidator.debug("Posting invalidation keys " + invKeys + 
                               " to " + server);

        StringBuffer dataBuffer = new StringBuffer();
        Iterator iter = invKeys.iterator();
        while (iter.hasNext()) {
            String remKey = (String)iter.next();
            dataBuffer.append(remKey);
            if (iter.hasNext()) {
                dataBuffer.append(DELIMITER);
            }
        }
                                
        try {
            HttpURLConnection connection = 
                (HttpURLConnection) server.openConnection();
            connection.setRequestMethod("POST");
            connection.setUseCaches(false);
            connection.setDoOutput(true);
            connection.setDoInput(true);
            connection.setRequestProperty("Content-type", "text/plain");
                    
            connection.connect();
                    
            // Send POST output.
            OutputStream out = connection.getOutputStream();
                    
            out.write(dataBuffer.toString().getBytes());
                    
            out.flush();
            out.close();
                    
            // 200 OK, 204 No Content
            int code = connection.getResponseCode();
            if ((code == 200) || (code == 204)) {
                success = true;
                CacheInvalidator.debug("Files deleted successfully from " +
                                       server);                    
            } else {
                CacheInvalidator.debug("Failed to delete files from " + 
                                       server +" " + code);
                success = false;
            }
        } catch (IOException ioe) {
            CacheInvalidator.debug("Failed to delete files from " +
                                   server);
        }
        return success;
    }

    /**
     * Kills the invalidator thread cleanly.
     */
    public void kill() {
        CacheInvalidator.debug("Killing invalidator thread " + server);
        synchronized (invalidationSet) {
            stop = true;
            interrupt();
        }
    }

    /**
     * Starts the invalidator thread.
     */
    public void run() {
        CacheInvalidator.debug("Invalidator thread started " + server);

        stop = false;

        while (!stop) {
            List remList = new LinkedList();
            synchronized (invalidationSet) {
                try {
                    if (invalidationSet.isEmpty()) {
                        invalidationSet.wait();
                    }
                } catch (InterruptedException e) { 
                    // no need to do anything
                }
            }

            if (!stop) {            
                try {
                    sleep(INVALIDATION_DELAY);
                } catch (InterruptedException ie) {
                    // no need to do anything
                }
            }

            synchronized (invalidationSet) {
                remList.addAll(invalidationSet);
                invalidationSet.clear();
            }
                
                
            if (post(remList)) {
                synchronized(invalidationSet) {
                    reset = false;
                }
            } else {
                synchronized(invalidationSet) {
                    invalidationSet.addAll(remList);
                }
            }
        }
        
        CacheInvalidator.debug("Invalidator thread stopped " + server);
    }
}
